/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/07/12
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.composer.exporter;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.List;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.CharEncoding;
import org.apache.commons.lang.SystemUtils;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;

import org.jiemamy.composer.ExportException;
import org.jiemamy.composer.Exporter;
import org.jiemamy.dialect.Dialect;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.sql.SqlStatement;

/**
 * モデルからSQLを構築するエクスポータ。
 * 
 * @author daisuke
 */
public class SqlExporter implements Exporter<SqlExportConfig> {
	
	/** ConfigKey: オーバーライトするかどうか (Boolean) */
	public static final String OVERWRITE = "overwrite";
	
	/** ConfigKey: 出力ファイル名 (String) */
	public static final String OUTPUT_FILE = "outputFile";
	
	/** ConfigKey: 出力データセット番号 (Integer) */
	public static final String DATA_SET_INDEX = "dataSetIndex";
	
	/** COnfigKey: DROP文を出力するかどうか (Boolean) */
	public static final String DROP = "drop";
	
	/** COnfigKey: DROP文を出力するかどうか (Boolean) */
	public static final String SCHEMA = "schema";
	

	public boolean exportModel(RootModel rootModel, SqlExportConfig config) throws ExportException {
		Validate.notNull(rootModel);
		Validate.notNull(config);
		Validate.notNull(config.getOutputFile());
		
		Writer writer = null;
		try {
			Dialect dialect = rootModel.findDialect();
			
			List<SqlStatement> statements = dialect.emitStatements(rootModel, config);
			
			File outputFile = config.getOutputFile();
			if (outputFile.exists()) {
				if (config.isOverwrite() == false) {
					return false;
				}
				if (outputFile.delete() == false) {
					throw new ExportException("Cannot delete file: " + outputFile.getAbsolutePath());
				}
			}
			
			writer = new OutputStreamWriter(new FileOutputStream(outputFile), CharEncoding.UTF_8);
			for (SqlStatement stmt : statements) {
				writer.write(stmt.toString());
				writer.write(SystemUtils.LINE_SEPARATOR);
			}
		} catch (IOException e) {
			throw new ExportException(e);
		} catch (ClassNotFoundException e) {
			throw new ExportException("Dialect not found: " + rootModel.getDialectClassName(), e);
		} catch (UnsupportedOperationException e) {
			throw new ExportException("This dialect does not support export SQL.", e);
		} finally {
			IOUtils.closeQuietly(writer);
		}
		return true;
	}
	
	public String getName() {
		return "SQL Exporter";
	}
	
	@Override
	public String toString() {
		return ToStringBuilder.reflectionToString(this, ToStringStyle.SHORT_PREFIX_STYLE);
	}
	
}
