/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/08/29
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.dialect.internal.impl;

import java.util.List;

import org.apache.commons.lang.Validate;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;

import org.jiemamy.JiemamyFactory;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.dialect.BuiltinDataTypeMold;
import org.jiemamy.dialect.DataTypeResolver;
import org.jiemamy.dialect.MoldManager;
import org.jiemamy.internal.model.sql.Keyword;
import org.jiemamy.internal.model.sql.Literal;
import org.jiemamy.internal.model.sql.Separator;
import org.jiemamy.model.datatype.BuiltinDataType;
import org.jiemamy.model.datatype.DataType;
import org.jiemamy.model.datatype.DataTypeCategory;
import org.jiemamy.model.datatype.adapter.PrecisionedDataTypeAdapter;
import org.jiemamy.model.datatype.adapter.SizedDataTypeAdapter;
import org.jiemamy.model.datatype.adapter.TimezonedDataTypeAdapter;
import org.jiemamy.model.sql.Token;
import org.jiemamy.utils.CollectionsUtil;

/**
 * SQL方言において、{@link DataType}と型名(String)間の解決を行うクラス。
 * 
 * @author daisuke
 */
public class DefaultDataTypeResolver implements DataTypeResolver {
	
	private final MoldManager moldManager;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param moldManager 鋳型マネージャ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public DefaultDataTypeResolver(MoldManager moldManager) {
		Validate.notNull(moldManager);
		this.moldManager = moldManager;
	}
	
	public List<Token> resolveDataType(DataType dataType, ReferenceResolver resolver) {
		List<Token> tokens = CollectionsUtil.newArrayList();
		BuiltinDataType builtinDataType = dataType.toBuiltinDataType(resolver);
		
		tokens.add(Keyword.of(moldManager.findDataTypeMold(builtinDataType).getName()));
		if (isSupport(SizedDataTypeAdapter.class, builtinDataType.getCategory())
				&& builtinDataType.hasAdapter(SizedDataTypeAdapter.class)) {
			Integer size = builtinDataType.getAdapter(SizedDataTypeAdapter.class).getSize();
			if (size != null) {
				tokens.add(Separator.LEFT_PAREN);
				tokens.add(Literal.of(size));
				tokens.add(Separator.RIGHT_PAREN);
			}
		}
		
		if (isSupport(PrecisionedDataTypeAdapter.class, builtinDataType.getCategory())
				&& builtinDataType.hasAdapter(PrecisionedDataTypeAdapter.class)) {
			PrecisionedDataTypeAdapter adapter = builtinDataType.getAdapter(PrecisionedDataTypeAdapter.class);
			Integer prec = adapter.getPrecision();
			Integer scale = adapter.getScale();
			if (prec != null) {
				tokens.add(Separator.LEFT_PAREN);
				tokens.add(Literal.of(prec));
				if (scale != null) {
					tokens.add(Separator.COMMA);
					tokens.add(Literal.of(scale));
				}
				tokens.add(Separator.RIGHT_PAREN);
			}
		}
		if (isSupport(TimezonedDataTypeAdapter.class, builtinDataType.getCategory())
				&& builtinDataType.hasAdapter(TimezonedDataTypeAdapter.class)) {
			TimezonedDataTypeAdapter adapter = builtinDataType.getAdapter(TimezonedDataTypeAdapter.class);
			Boolean withTimezone = adapter.isWithTimezone();
			if (withTimezone != null) {
				tokens.add(withTimezone ? Keyword.WITH : Keyword.WITHOUT);
				tokens.add(Keyword.TIMEZONE);
			}
		}
		return tokens;
	}
	
	public BuiltinDataType resolveDataType(JiemamyFactory factory, int sqlType, String typeName) {
		Validate.notNull(factory);
		
		DataTypeCategory category;
		try {
			category = DataTypeCategory.valueOf(typeName);
		} catch (IllegalArgumentException e) {
			category = DataTypeCategory.fromSqlType(sqlType);
		}
		BuiltinDataTypeMold mold = moldManager.findDataTypeMold(category, typeName);
		
		return factory.newDataType(mold);
	}
	
	@Override
	public String toString() {
		return ToStringBuilder.reflectionToString(this, ToStringStyle.SHORT_PREFIX_STYLE);
	}
	
	private boolean isSupport(Class<?> adapterClass, DataTypeCategory dataTypeCategory) {
		List<BuiltinDataTypeMold> allDataTypes = moldManager.getTypeList();
		for (BuiltinDataTypeMold mold : allDataTypes) {
			if (mold.getCategory() == dataTypeCategory && mold.getSupportedAdapterClasses().contains(adapterClass)) {
				return true;
			}
		}
		return false;
	}
}
