/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/03/31
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.dialect.internal.impl;

import org.apache.commons.lang.Validate;

import org.jiemamy.dialect.ConnectionAdvisor;
import org.jiemamy.dialect.DataTypeResolver;
import org.jiemamy.dialect.MoldManager;
import org.jiemamy.dialect.internal.DialectInitializer;
import org.jiemamy.dialect.internal.EntityImportVisitor;
import org.jiemamy.dialect.internal.ForeignKeyImportVisitor;
import org.jiemamy.dialect.internal.SqlEmitter;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.datatype.DataType;

/**
 * {@link DialectInitializer}のデフォルト実装クラス。
 * 
 * @author daisuke
 */
public class DefaultDialectInitializer implements DialectInitializer {
	
	/** データ型鋳型の管理クラス */
	private MoldManager moldManager;
	
	/** データ型（{@link DataType}）を中心とした各種解決を行うロジッククラス */
	private DataTypeResolver dataTypeResolver;
	
	/** DBメタデータの読み出し前後に実行するロジッククラス */
	private ConnectionAdvisor connectionAdvisor;
	
	/** {@link RootModel}からSQLのリストを生成するエミッタ */
	private SqlEmitter sqlEmitter;
	
	/** エンティティメタ情報から{@link RootModel}にインポートを行うビジター */
	private EntityImportVisitor entityImportVisitor;
	
	/** 外部キーメタ情報から{@link RootModel}にインポートを行うビジター */
	private ForeignKeyImportVisitor foreignKeyImportVisitor;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param moldManager データ型鋳型の管理クラス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public DefaultDialectInitializer(MoldManager moldManager) {
		Validate.notNull(moldManager);
		this.moldManager = moldManager;
		dataTypeResolver = new DefaultDataTypeResolver(moldManager);
		connectionAdvisor = new DefaultConnectionAdvisor();
		sqlEmitter = new DefaultSqlEmitter(dataTypeResolver);
		entityImportVisitor = new DefaultEntityImportVisitor();
		foreignKeyImportVisitor = new DefaultForeignKeyImportVisitor();
	}
	
	public ConnectionAdvisor getConnectionAdvisor() {
		return connectionAdvisor;
	}
	
	public DataTypeResolver getDataTypeResolver() {
		return dataTypeResolver;
	}
	
	public EntityImportVisitor getEntityImportVisitor() {
		return entityImportVisitor;
	}
	
	public ForeignKeyImportVisitor getForeignKeyImportVisitor() {
		return foreignKeyImportVisitor;
	}
	
	public MoldManager getMoldManager() {
		return moldManager;
	}
	
	public SqlEmitter getSqlEmitter() {
		return sqlEmitter;
	}
	
	/**
	 * DBメタデータの読み出し前後に実行するロジッククラスを設定する。
	 * 
	 * @param connectionAdvisor DBメタデータの読み出し前後に実行するロジッククラス
	 */
	public void setConnectionAdvisor(ConnectionAdvisor connectionAdvisor) {
		this.connectionAdvisor = connectionAdvisor;
	}
	
	/**
	 * データ型（{@link DataType}）を中心とした各種解決を行うロジッククラスを設定する。
	 * 
	 * @param dataTypeResolver データ型（{@link DataType}）を中心とした各種解決を行うロジッククラス
	 */
	public void setDataTypeResolver(DataTypeResolver dataTypeResolver) {
		this.dataTypeResolver = dataTypeResolver;
	}
	
	/**
	 * エンティティメタ情報から{@link RootModel}にインポートを行うビジターを設定する。
	 * 
	 * @param entityImportVisitor エンティティメタ情報から{@link RootModel}にインポートを行うビジター
	 */
	public void setEntityImportVisitor(EntityImportVisitor entityImportVisitor) {
		this.entityImportVisitor = entityImportVisitor;
	}
	
	/**
	 * 外部キーメタ情報から{@link RootModel}にインポートを行うビジターを設定する。
	 * 
	 * @param foreignKeyImportVisitor 外部キーメタ情報から{@link RootModel}にインポートを行うビジター
	 */
	public void setForeignKeyImportVisitor(ForeignKeyImportVisitor foreignKeyImportVisitor) {
		this.foreignKeyImportVisitor = foreignKeyImportVisitor;
	}
	
	/**
	 * データ型鋳型の管理クラスを設定する。
	 * 
	 * @param moldManager データ型鋳型の管理クラス
	 */
	public void setMoldManager(MoldManager moldManager) {
		this.moldManager = moldManager;
	}
	
	/**
	 * {@link RootModel}からSQLのリストを生成するエミッタを設定する。
	 * 
	 * @param sqlEmitter {@link RootModel}からSQLのリストを生成するエミッタ
	 */
	public void setSqlEmitter(SqlEmitter sqlEmitter) {
		this.sqlEmitter = sqlEmitter;
	}
	
}
