/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/12/29
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal;

import java.util.Collection;
import java.util.Set;
import java.util.UUID;

import org.apache.commons.lang.Validate;

import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.JiemamyImplementation;
import org.jiemamy.OptionalSpec;
import org.jiemamy.dialect.BuiltinDataTypeMold;
import org.jiemamy.exception.IllegalImplementationException;
import org.jiemamy.experimental.builder.JiemamyBuilder;
import org.jiemamy.facade.JiemamyFacade;
import org.jiemamy.internal.enhance.FactoryExtension;
import org.jiemamy.internal.model.datatype.BuiltinDataTypeImpl;
import org.jiemamy.internal.serializer.JiemamyDomSerializer;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.datatype.BuiltinDataType;
import org.jiemamy.model.datatype.DataType;
import org.jiemamy.model.datatype.DataTypeMold;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.datatype.DomainRef;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.serializer.JiemamySerializer;
import org.jiemamy.utils.ValidateUtil;

/**
 * {@link JiemamyFactory}のArtemis用実装。
 * 
 * @author daisuke
 */
public class ArtemisJiemamyFactory implements JiemamyFactory {
	
	/** Jiemamy実装（モデルインスタンス生成戦略） */
	private JiemamyImplementation implementation;
	
	/** このファクトリが生成する{@link JiemamyElement}が属するインスタンス空間 */
	private Jiemamy jiemamy;
	
	/** このファクトリが生成した、唯一の{@link RootModel} */
	private RootModel rootModel;
	

	public JiemamyImplementation getImplementation() {
		checkInitialized();
		return implementation;
	}
	
	public synchronized RootModel getRootModel() {
		if (hasRootModel() == false) {
			rootModel = newRootModel();
		}
		return rootModel;
	}
	
	public synchronized RootModel getRootModel(UUID id) {
		if (hasRootModel()) {
			if (getRootModel().getId().equals(id) == false) {
				throw new IllegalStateException();
			}
			rootModel = getRootModel();
		} else {
			rootModel = newRootModel(id);
		}
		return rootModel;
	}
	
	public JiemamySerializer getSerializer() {
		checkInitialized();
		return new JiemamyDomSerializer(jiemamy);
	}
	
	/**
	 * {@inheritDoc}
	 * この実装では、毎回異なった{@link Set}インスタンスを返す。
	 */
	public Set<OptionalSpec> getSupportedSpecs() {
		checkInitialized();
		return implementation.getSupportedSpecs();
	}
	
	public synchronized boolean hasRootModel() {
		return rootModel != null;
	}
	
	public <T>T newAdapter(Class<T> clazz) {
		Validate.notNull(clazz);
		checkInitialized();
		T adapter = new FactoryExtension<T>().newAdapter(clazz);
		if (adapter == null) {
			throw new IllegalArgumentException();
		}
		return adapter;
	}
	
	public <M extends JiemamyElement, T extends JiemamyBuilder<M>>T newBuilder(Class<T> clazz) {
		Validate.notNull(clazz);
		checkInitialized();
		T builder = new FactoryExtension<T>().newBuilder(clazz, jiemamy.getFactory());
		if (builder == null) {
			throw new IllegalArgumentException();
		}
		return builder;
	}
	
	@SuppressWarnings("unchecked")
	public <T extends DataType>T newDataType(DataTypeMold<T> mold) {
		Validate.notNull(mold);
		if (mold instanceof DomainModel) {
			DomainModel domainModel = (DomainModel) mold;
			return (T) newReference(domainModel);
		} else if (mold instanceof BuiltinDataTypeMold) {
			BuiltinDataTypeMold builtinDataTypeMold = (BuiltinDataTypeMold) mold;
			return (T) newBuiltinDataType(builtinDataTypeMold);
		}
		return null;
	}
	
	@SuppressWarnings("unchecked")
	public <T extends JiemamyFacade>T newFacade(Class<T> clazz) {
		Validate.notNull(clazz);
		checkInitialized();
		return (T) new JiemamyFacadeImpl(jiemamy);
	}
	
	public synchronized <T extends JiemamyElement>T newModel(Class<T> clazz) {
		checkInitialized();
		if (hasRootModel() == true && clazz == RootModel.class) {
			throw new IllegalStateException();
		}
		T model = new FactoryExtension<T>().newInstance(clazz, jiemamy);
		if (model == null) {
			throw new IllegalArgumentException();
		}
		if (clazz == RootModel.class) {
			rootModel = (RootModel) model;
		}
		implementation.getReferenceResolver().add(model);
		return model;
	}
	
	public synchronized <T extends JiemamyElement>T newModel(Class<T> clazz, UUID id) {
		checkInitialized();
		if (hasRootModel() == true && clazz == RootModel.class) {
			throw new IllegalStateException();
		}
		T model = new FactoryExtension<T>().newInstance(clazz, jiemamy, id);
		if (model == null) {
			throw new IllegalArgumentException(clazz.getName());
		}
		if (clazz == RootModel.class) {
			rootModel = (RootModel) model;
		}
		implementation.getReferenceResolver().add(model);
		assert model.getJiemamy() != null;
		return model;
	}
	
	public ColumnRef newReference(ColumnModel columnModel) {
		return new FactoryExtension<ColumnRef>().newReference(columnModel);
	}
	
	public DomainRef newReference(DomainModel domainModel) {
		return new FactoryExtension<DomainRef>().newReference(domainModel);
	}
	
	public TableRef newReference(TableModel tableModel) {
		return new FactoryExtension<TableRef>().newReference(tableModel);
	}
	
	public void setImplementation(JiemamyImplementation implementation) {
		Validate.notNull(implementation);
		ValidateUtil.injectionSetter(this.implementation);
		this.implementation = implementation;
	}
	
	public void setJiemamy(Jiemamy jiemamy) {
		Validate.notNull(jiemamy);
		ValidateUtil.injectionSetter(this.jiemamy);
		this.jiemamy = jiemamy;
	}
	
	RootModel newRootModel() {
		RootModel model = newModel(RootModel.class);
		assert model != null;
		assert model.getJiemamy() != null;
		return model;
	}
	
	RootModel newRootModel(UUID id) {
		RootModel model = newModel(RootModel.class, id);
		assert model != null;
		assert model.getJiemamy() != null;
		return model;
	}
	
	private void checkInitialized() {
		if (jiemamy == null) {
			throw new IllegalStateException("jiemamy is null.");
		}
		if (implementation == null) {
			throw new IllegalStateException("implementation is null.");
		}
	}
	
	private BuiltinDataType newBuiltinDataType(BuiltinDataTypeMold mold) {
		Validate.notNull(mold);
		checkInitialized();
		BuiltinDataTypeImpl dataType = new BuiltinDataTypeImpl(mold.getCategory(), mold.getName());
		Collection<Class<?>> supportedAdapterClasses = mold.getSupportedAdapterClasses();
		for (Class<?> supportedAdapterClass : supportedAdapterClasses) {
			try {
				Object adapter = supportedAdapterClass.newInstance();
				dataType.registerAdapter(adapter);
			} catch (InstantiationException e) {
				throw new IllegalImplementationException(supportedAdapterClass);
			} catch (IllegalAccessException e) {
				throw new IllegalImplementationException(supportedAdapterClass);
			}
		}
		return dataType;
	}
	
}
