/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/02/22
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal;

import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.jiemamy.EventBroker;
import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.JiemamyProperty;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.JiemamyProperty.ColumnProperty;
import org.jiemamy.dialect.BuiltinDataTypeMold;
import org.jiemamy.dialect.Dialect;
import org.jiemamy.dialect.generic.GenericDialect;
import org.jiemamy.editcommand.Command;
import org.jiemamy.facade.JiemamyFacade;
import org.jiemamy.facade.SavePoint;
import org.jiemamy.internal.editcommand.AddAttributeCommand;
import org.jiemamy.internal.editcommand.AddColumnToColumnRefListCommand;
import org.jiemamy.internal.editcommand.AddDataSetToRootCommand;
import org.jiemamy.internal.editcommand.AddDomainToRootCommand;
import org.jiemamy.internal.editcommand.AddEntityToRootCommand;
import org.jiemamy.internal.editcommand.AddIndexColumnCommand;
import org.jiemamy.internal.editcommand.AddIndexCommand;
import org.jiemamy.internal.editcommand.AddRecordsToDataSetCommand;
import org.jiemamy.internal.editcommand.ModifyModelPropertyCommand;
import org.jiemamy.internal.editcommand.RemoveAttributeCommand;
import org.jiemamy.internal.editcommand.RemoveColumnFromColumnRefListCommand;
import org.jiemamy.internal.editcommand.RemoveDataSetFromRootCommand;
import org.jiemamy.internal.editcommand.RemoveDomainFromRootCommand;
import org.jiemamy.internal.editcommand.RemoveEntityFromRootCommand;
import org.jiemamy.internal.editcommand.RemoveIndexColumnCommand;
import org.jiemamy.internal.editcommand.RemoveIndexCommand;
import org.jiemamy.internal.editcommand.RemoveRecordsFromDataSetCommand;
import org.jiemamy.internal.editcommand.SetRecordValueCommand;
import org.jiemamy.internal.editcommand.SwapListElementCommand;
import org.jiemamy.internal.editcommand.UpdateForeignKeyMappingCommand;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.model.attribute.constraint.LocalKeyConstraint;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.BuiltinDataType;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.datatype.DomainRef;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.utils.ArrayEssentialStack;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.utils.EssentialStack;
import org.jiemamy.utils.EssentialStacks;

/**
 * Jiemamyに対するファサードの実装クラス。
 * 
 * @author daisuke
 */
public class JiemamyFacadeImpl implements JiemamyFacade {
	
	private static Logger logger = LoggerFactory.getLogger(JiemamyFacadeImpl.class);
	
	/** ファサードによる操作の対象となるコンテキスト */
	protected final Jiemamy jiemamy;
	
	/** このファサードが発行したセーブポイントの集合 */
	private Collection<SavePoint> publishedSavePoints = CollectionsUtil.newArrayList();
	
	/** UNDOスタック */
	protected EssentialStack<Command> undoStack = new ArrayEssentialStack<Command>();
	
	/** イベントブローカ */
	protected EventBroker eventBroker;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param jiemamy ファサードによる操作の対象となるコンテキスト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public JiemamyFacadeImpl(Jiemamy jiemamy) {
		Validate.notNull(jiemamy);
		
		this.jiemamy = jiemamy;
		eventBroker = jiemamy.getEventBroker();
	}
	
	public void addAttribute(TableModel tableModel, AttributeModel attributeModel) {
		Validate.notNull(tableModel);
		Validate.notNull(attributeModel);
		if (tableModel.getAttributes().contains(attributeModel)) {
			return;
		}
		
		if (attributeModel instanceof KeyConstraint) {
			checkKeyConstraint(tableModel, (KeyConstraint) attributeModel);
		}
		
		AddAttributeCommand command = new AddAttributeCommand(eventBroker, tableModel, attributeModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addAttribute(TableModel tableModel, int index, AttributeModel attributeModel) {
		Validate.notNull(tableModel);
		Validate.notNull(attributeModel);
		if (tableModel.getAttributes().contains(attributeModel)) {
			return;
		}
		
		if (attributeModel instanceof KeyConstraint) {
			checkKeyConstraint(tableModel, (KeyConstraint) attributeModel);
		}
		
		AddAttributeCommand command = new AddAttributeCommand(eventBroker, tableModel, index, attributeModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addDataSet(DataSetModel dataSetModel) {
		Validate.notNull(dataSetModel);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		if (rootModel.getDataSets().contains(dataSetModel)) {
			return;
		}
		
		AddDataSetToRootCommand command = new AddDataSetToRootCommand(eventBroker, rootModel, dataSetModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addDataSet(DataSetModel dataSetModel, int index) {
		Validate.notNull(dataSetModel);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		if (rootModel.getDataSets().contains(dataSetModel)) {
			return;
		}
		
		AddDataSetToRootCommand command = new AddDataSetToRootCommand(eventBroker, rootModel, dataSetModel, index);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addDomain(DomainModel domainModel) {
		Validate.notNull(domainModel);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		if (rootModel.getDomains().contains(domainModel)) {
			return;
		}
		
		AddDomainToRootCommand command = new AddDomainToRootCommand(eventBroker, rootModel, domainModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addDomain(DomainModel domainModel, int index) {
		Validate.notNull(domainModel);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		if (rootModel.getDomains().contains(domainModel)) {
			return;
		}
		
		AddDomainToRootCommand command = new AddDomainToRootCommand(eventBroker, rootModel, domainModel, index);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addEntity(EntityModel entityModel) {
		Validate.notNull(entityModel);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		if (rootModel.getEntities().contains(entityModel)) {
			return;
		}
		
		AddEntityToRootCommand command = new AddEntityToRootCommand(eventBroker, rootModel, entityModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addIndex(TableModel tableModel, IndexModel indexModel) {
		Validate.notNull(tableModel);
		Validate.notNull(indexModel);
		if (tableModel.getIndexes().contains(indexModel)) {
			return;
		}
		
		AddIndexCommand command = new AddIndexCommand(eventBroker, tableModel, indexModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addIndex(TableModel tableModel, int index, IndexModel indexModel) {
		Validate.notNull(tableModel);
		Validate.notNull(indexModel);
		if (tableModel.getIndexes().contains(indexModel)) {
			return;
		}
		
		AddIndexCommand command = new AddIndexCommand(eventBroker, tableModel, index, indexModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addIndexColumn(IndexModel indexModel, IndexColumnModel indexColumnModel) {
		Validate.notNull(indexColumnModel);
		Validate.notNull(indexColumnModel);
		if (indexModel.getIndexColumns().contains(indexColumnModel)) {
			return;
		}
		
		AddIndexColumnCommand command = new AddIndexColumnCommand(eventBroker, indexModel, indexColumnModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addIndexColumn(IndexModel indexModel, int index, IndexColumnModel indexColumnModel) {
		Validate.notNull(indexColumnModel);
		Validate.notNull(indexColumnModel);
		if (indexModel.getIndexColumns().contains(indexColumnModel)) {
			return;
		}
		
		AddIndexColumnCommand command = new AddIndexColumnCommand(eventBroker, indexModel, index, indexColumnModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addKeyColumn(KeyConstraint keyConstraint, ColumnModel columnModel) {
		Validate.notNull(keyConstraint);
		Validate.notNull(columnModel);
		// 既に指定したカラムがキーカラムの要素であった場合は何もしない
		ColumnRef columnRef = columnModel.getJiemamy().getFactory().newReference(columnModel);
		if (keyConstraint.getKeyColumns().contains(columnRef)) {
			return;
		}
		
		AddColumnToColumnRefListCommand command =
				new AddColumnToColumnRefListCommand(eventBroker, keyConstraint, keyConstraint.getKeyColumns(),
						columnModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addRecords(DataSetModel dataSetModel, TableModel tableModel) {
		Validate.notNull(dataSetModel);
		Validate.notNull(tableModel);
		
		Set<TableRef> keySet = dataSetModel.getRecords().keySet();
		if (keySet != null) {
			for (TableRef tableRef : keySet) {
				if (jiemamy.getReferenceResolver().canResolve(tableRef)
						&& jiemamy.getReferenceResolver().resolve(tableRef) == tableModel) {
					return;
				}
			}
		}
		
		Command command = new AddRecordsToDataSetCommand(eventBroker, dataSetModel, tableModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public <T extends JiemamyElement>void changeModelProperty(T element, JiemamyProperty<T> property, Object newValue) {
		Command command = new ModifyModelPropertyCommand<T>(eventBroker, element, property, newValue);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeAttribute(TableModel tableModel, AttributeModel attributeModel) {
		Validate.notNull(tableModel);
		Validate.notNull(attributeModel);
		// ローカルキーを削除する場合、そのキーを参照する外部キーも同時に削除する
		if (attributeModel instanceof LocalKeyConstraint) {
			LocalKeyConstraint localKeyConstraint = (LocalKeyConstraint) attributeModel;
			TableModel definedTable = localKeyConstraint.findDeclaringTable();
			Collection<EntityModel> dependentEntities = definedTable.findSubEntities(false);
			for (EntityModel entityModel : dependentEntities) {
				if (entityModel instanceof TableModel) {
					TableModel dependentTable = (TableModel) entityModel;
					List<ForeignKey> foreignKeys = dependentTable.findAttributes(ForeignKey.class);
					for (ForeignKey foreignKey : foreignKeys) {
						if (foreignKey.findReferencedKeyConstraint() == null) {
							removeAttribute(dependentTable, foreignKey);
						}
					}
				}
			}
		}
		
		// カラムを削除する場合、全てのキーに含まれる該当カラムへの参照も同時に削除する
		if (attributeModel instanceof ColumnModel) {
			ColumnModel columnModel = (ColumnModel) attributeModel;
			List<KeyConstraint> keys = tableModel.findAttributes(KeyConstraint.class, true);
			for (KeyConstraint keyConstraint : keys) {
				removeKeyColumn(keyConstraint, columnModel);
			}
		}
		
		RemoveAttributeCommand command = new RemoveAttributeCommand(eventBroker, tableModel, attributeModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeDataSet(DataSetModel dataSetModel) {
		Validate.notNull(dataSetModel);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		if (rootModel.getDataSets().contains(dataSetModel) == false) {
			return;
		}
		
		RemoveDataSetFromRootCommand command = new RemoveDataSetFromRootCommand(eventBroker, rootModel, dataSetModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeDomain(DomainModel domainModel) {
		Validate.notNull(domainModel);
		JiemamyFactory factory = jiemamy.getFactory();
		RootModel rootModel = factory.getRootModel();
		if (rootModel.getDomains().contains(domainModel) == false) {
			return;
		}
		
		// 該当ドメインへの参照は、ドメインに設定されていた型情報に書き換える
		DomainRef domainRef = factory.newReference(domainModel);
		Dialect dialect;
		try {
			dialect = rootModel.findDialect();
		} catch (ClassNotFoundException e) {
			logger.warn("Dialectのロスト", e);
			dialect = new GenericDialect();
		}
		Collection<TableModel> tableModels = rootModel.findEntities(TableModel.class);
		for (TableModel tableModel : tableModels) {
			List<ColumnModel> columns = tableModel.findColumns();
			for (ColumnModel columnModel : columns) {
				if (columnModel.getDataType().equals(domainRef)) {
					BuiltinDataType domainDataType = domainModel.getDataType();
					BuiltinDataTypeMold mold = dialect.findDataTypeMold(domainDataType);
					BuiltinDataType newDataType = factory.newDataType(mold);
					
					List<Object> domainDataTypeAdapters = domainDataType.getAdapters();
					for (Object domainDtaTypeAdapter : domainDataTypeAdapters) {
						Class<?> adapterClass = domainDtaTypeAdapter.getClass();
						if (newDataType.hasAdapter(adapterClass)) {
							Object newDataTypeAdapter = newDataType.getAdapter(adapterClass);
							try {
								BeanUtils.copyProperties(newDataTypeAdapter, domainDtaTypeAdapter);
							} catch (IllegalAccessException e) {
								logger.error("アダプタのコピーに失敗", e);
							} catch (InvocationTargetException e) {
								logger.error("アダプタのコピーに失敗", e);
							}
						}
					}
					changeModelProperty(columnModel, ColumnProperty.dataType, newDataType);
				}
			}
		}
		
		RemoveDomainFromRootCommand command = new RemoveDomainFromRootCommand(eventBroker, rootModel, domainModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeEntity(EntityModel entityModel) {
		Validate.notNull(entityModel);
		JiemamyFactory factory = jiemamy.getFactory();
		RootModel rootModel = factory.getRootModel();
		if (rootModel.getEntities().contains(entityModel) == false) {
			return;
		}
		
		if (entityModel instanceof TableModel) {
			// 該当エンティティを参照する外部キーも同時に削除する
			TableModel tableModel = (TableModel) entityModel;
			Collection<EntityModel> sourceEntities = tableModel.findSubEntities(false);
			for (EntityModel sourceEntity : sourceEntities) {
				if (sourceEntity instanceof TableModel) {
					TableModel sourceTable = (TableModel) sourceEntity;
					List<ForeignKey> foreignKeys = sourceTable.findAttributes(ForeignKey.class);
					for (ForeignKey foreignKey : foreignKeys) {
						if (foreignKey.findReferencedEntity() == tableModel) {
							removeAttribute(sourceTable, foreignKey);
						}
					}
				}
			}
			
			// DataSetModelが持つテーブルへの参照も同時に削除する
			List<DataSetModel> dataSets = rootModel.getDataSets();
			for (DataSetModel dataSetModel : dataSets) {
				removeRecords(dataSetModel, tableModel);
			}
		}
		
		RemoveEntityFromRootCommand command = new RemoveEntityFromRootCommand(eventBroker, rootModel, entityModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeIndex(TableModel tableModel, IndexModel indexModel) {
		Validate.notNull(tableModel);
		Validate.notNull(indexModel);
		if (tableModel.getIndexes().contains(indexModel) == false) {
			return;
		}
		
		RemoveIndexCommand command = new RemoveIndexCommand(eventBroker, tableModel, indexModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeIndexColumn(IndexModel indexModel, IndexColumnModel indexColumnModel) {
		Validate.notNull(indexModel);
		Validate.notNull(indexColumnModel);
		if (indexModel.getIndexColumns().contains(indexColumnModel) == false) {
			return;
		}
		
		RemoveIndexColumnCommand command = new RemoveIndexColumnCommand(eventBroker, indexModel, indexColumnModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeKeyColumn(KeyConstraint keyConstraint, ColumnModel columnModel) {
		Validate.notNull(keyConstraint);
		Validate.notNull(columnModel);
		// キーカラム要素が、指定したカラムへの参照を保持していなかった場合は、何もしない
		JiemamyFactory factory = jiemamy.getFactory();
		int index = keyConstraint.getKeyColumns().indexOf(factory.newReference(columnModel));
		if (index == ArrayUtils.INDEX_NOT_FOUND) {
			return;
		}
		
		RemoveColumnFromColumnRefListCommand command =
				new RemoveColumnFromColumnRefListCommand(eventBroker, keyConstraint, keyConstraint.getKeyColumns(),
						columnModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
		
		// 削除の結果、キーカラムの数が0になった場合、キー制約自体を所属テーブルから削除する
		if (keyConstraint.getKeyColumns().size() == 0) {
			TableModel definedTable = keyConstraint.findDeclaringTable();
			if (definedTable != null) {
				removeAttribute(definedTable, keyConstraint);
			}
		}
		
		// 外部キーであった場合は、対応する参照カラムも削除する
		if (keyConstraint instanceof ForeignKey) {
			ForeignKey foreignKey = (ForeignKey) keyConstraint;
			ColumnRef columnRef = foreignKey.getReferenceColumns().get(index);
			ColumnModel columnModel2 = jiemamy.getReferenceResolver().resolve(columnRef);
			RemoveColumnFromColumnRefListCommand command2 =
					new RemoveColumnFromColumnRefListCommand(eventBroker, foreignKey, foreignKey.getReferenceColumns(),
							columnModel2);
			command2.execute();
			undoStack.push(command2.getNegateCommand());
		}
	}
	
	public void removeRecords(DataSetModel dataSetModel, TableModel tableModel) {
		Validate.notNull(dataSetModel);
		Validate.notNull(tableModel);
		if (dataSetModel.getRecords().containsKey(jiemamy.getFactory().newReference(tableModel)) == false) {
			return;
		}
		
		Command command = new RemoveRecordsFromDataSetCommand(eventBroker, dataSetModel, tableModel);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void rollback(SavePoint savePoint) {
		Validate.notNull(savePoint);
		if (publishedSavePoints.contains(savePoint) == false) {
			throw new IllegalArgumentException();
		}
		
		// this.undoStack(現状スタック)が ABCPQR, savePointStack(セーブ時スタック)が ABCXYZ だったとすると…
		
		EssentialStack<Command> savePointStack = ((SavePointImpl) savePoint).getUndoStackSnapshot();
		
		// intersect = ABC (先頭共通部分の抽出)
		EssentialStack<Command> intersect = EssentialStacks.intersection(undoStack, savePointStack);
		
		// undoStackMinus = PQR (引き算)
		EssentialStack<Command> undoStackMinus = EssentialStacks.minus(undoStack, intersect);
		
		// RQPの順番（pop順）でundoコマンドを実行し、現状を ABC 状態まで退行させる。
		while (undoStackMinus.isEmpty() == false) {
			Command command = undoStackMinus.pop();
			command.execute();
		}
		
		// savePointStackMinus = XYZ (引き算)
		EssentialStack<Command> savePointStackMinus = EssentialStacks.minus(savePointStack, intersect);
		
		// XYZの順番（foreach順）でredoコマンド（スタックから得られたコマンドの逆コマンド）を実行し、savePoint状態を再現する。
		for (Command negateCommand : savePointStackMinus) {
			Command command = negateCommand.getNegateCommand();
			command.execute();
		}
		
		// 現状スタックをセーブ時の状況に直しておく。
		undoStack = savePointStack;
	}
	
	public SavePoint save() {
		EssentialStack<Command> undoCopy = new ArrayEssentialStack<Command>(undoStack);
		SavePoint sp = new SavePointImpl(undoCopy);
		publishedSavePoints.add(sp);
		return sp;
	}
	
	public void setRecordValue(RecordModel recordModel, ColumnModel columnModel, String value) {
		Command command = new SetRecordValueCommand(eventBroker, recordModel, columnModel, value);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void swapListElement(JiemamyElement targetElement, List<?> list, int subjectIndex, int objectIndex) {
		Command command = new SwapListElementCommand(eventBroker, targetElement, list, subjectIndex, objectIndex);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void updateForeignKeyMapping(ForeignKey foreignKey, List<ColumnRef> keyColumns,
			List<ColumnRef> referenceColumns) {
		Validate.isTrue(keyColumns.size() == referenceColumns.size());
		
		Command command = new UpdateForeignKeyMappingCommand(eventBroker, foreignKey, keyColumns, referenceColumns);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	private void checkKeyConstraint(TableModel tableModel, KeyConstraint keyConstraint) {
		ReferenceResolver referenceResolver = jiemamy.getReferenceResolver();
		List<ColumnRef> keyColumns = keyConstraint.getKeyColumns();
		for (ColumnRef columnRef : keyColumns) {
			ColumnModel columnModel = referenceResolver.resolve(columnRef);
			if (tableModel.getAttributes().contains(columnModel) == false) {
				throw new IllegalArgumentException();
			}
		}
	}
}
