/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/18
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.UUID;
import java.util.Map.Entry;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.jiemamy.ReferenceResolveException;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.model.ElementReference;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.utils.LogMarker;

/**
 * モデルの参照オブジェクトからモデルの実体を解決するリゾルバ。
 * 
 * @author daisuke
 */
public class ReferenceResolverImpl implements ReferenceResolver {
	
	static Logger logger = LoggerFactory.getLogger(ReferenceResolverImpl.class);
	
	/**
	 * {@link JiemamyElement}のidと{@link JiemamyElement}の実体をマッピングする{@link Map}
	 * 
	 * <p>FIXME マッピングが登録されることはあっても、リリースされることがない。
	 * モデルを大量に生成した場合、{@link RootModel}からたどることのないモデルであっても、このMapが参照を握ってヒープが溢れる。
	 * {@link ReferenceResolverImplTest#test00_メモリリークテスト()}及び CORE-105 を参照。<p>
	 */
	private static Map<UUID, JiemamyElement> elementMapping = new HashMap<UUID, JiemamyElement>();
	

	/**
	 * 指定された{@link UUID}が、指定されたElementから見た関係者かどうかを深く探索する。
	 * 
	 * <p>CollectionやMapはその要素を対象にさらに深く検索するが、ElementReferenceの中は一階層だけしか見ない。</p>
	 * <p>イベントの通知対象かどうかを判断するために導入した。</p>
	 * 
	 * @param element 基準要素
	 * @param id モデルID
	 * @return 関係があるなら{@code true}、そうでないなら{@code false}
	 */
	public static boolean isDescendFromElement(JiemamyElement element, UUID id) {
		boolean traceEnabled = logger.isTraceEnabled();
		if (traceEnabled) {
			logger.trace(LogMarker.LIFECYCLE, "isDescendFromElement() uuid=" + id + ", " + element);
		}
		if (element == null) {
			return false;
		}
		if (element.getId().equals(id)) {
			return true;
		}
		try {
			// 調査対象のモデルのフィールドを全部調べる。
			@SuppressWarnings("unchecked")
			Map<String, Object> description = PropertyUtils.describe(element);
			Iterator<Entry<String, Object>> i = description.entrySet().iterator();
			while (i.hasNext()) {
				Entry<String, Object> next = i.next();
				Object value = next.getValue();
				logger.trace(String.format("  %s@%s.%s=%s", new Object[] {
					element.getClass().getSimpleName(),
					Integer.toHexString(element.hashCode()),
					next.getKey(),
					next.getValue()
				}));
				if (isDescendFromFieldValue(value, id)) {
					return true;
				}
			}
			// 登録されたAdapterもフィールドと同じ扱いで調査対象にする。
			Iterator<Object> adapters = element.getAdapters().iterator();
			while (adapters.hasNext()) {
				Object value = adapters.next();
				logger.trace(String.format("  %s@%s.adapter=%s", new Object[] {
					element.getClass().getSimpleName(),
					Integer.toHexString(element.hashCode()),
					ToStringBuilder.reflectionToString(value)
				}));
				if (isDescendFromFieldValue(value, id)) {
					return true;
				}
			}
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * {@link #isDescendFromElement(JiemamyElement, UUID)}のヘルパーメソッド。
	 * 
	 * @param fieldValue 調査する対象のFieldの値
	 * @param id モデルID
	 * @return 関係があるなら{@code true}、そうでないなら{@code false}
	 */
	private static boolean isDescendFromFieldValue(Object fieldValue, UUID id) {
		if (fieldValue == null) {
			return false;
		}
		if (fieldValue instanceof JiemamyElement) {
			JiemamyElement e = (JiemamyElement) fieldValue;
			if (id.equals(e.getId())) {
				return true;
			} else {
				// JiemamyElementで、検索対象でなかった場合は再起して調査する。
				if (isDescendFromElement(e, id)) {
					return true;
				}
			}
		} else if (fieldValue instanceof ElementReference) {
			ElementReference<?> elementReferenceValue = (ElementReference<?>) fieldValue;
			// ElementReferenceだった場合は一階層だけ調査する。
			if (id.equals(elementReferenceValue.getReferenceId())) {
				return true;
			}
		} else if (fieldValue instanceof Iterable) {
			Iterable<?> iterable = (Iterable<?>) fieldValue;
			// Iterableだった場合は要素に対して再起して調査する。
			for (Object element : iterable) {
				if (isDescendFromFieldValue(element, id)) {
					return true;
				}
			}
		} else if (fieldValue instanceof Map) {
			Map<?, ?> map = (Map<?, ?>) fieldValue;
			// Mapだった場合は要素に対して再起して調査する。
			for (Object element : map.values()) {
				if (isDescendFromFieldValue(element, id)) {
					return true;
				}
			}
		}
		return false;
	}
	
	public void add(JiemamyElement model) {
		assert elementMapping != null;
		Validate.notNull(model);
		UUID id = model.getId();
		elementMapping.put(id, model);
	}
	
	public boolean canResolve(ElementReference<?> reference) {
		assert elementMapping != null;
		return elementMapping.containsKey(reference.getReferenceId());
	}
	
	@SuppressWarnings("unchecked")
	public <T extends JiemamyElement>T resolve(ElementReference<T> reference) {
		assert elementMapping != null;
		Validate.notNull(reference);
		
		if (canResolve(reference) == false) {
			throw new ReferenceResolveException(reference);
		}
		
		return (T) elementMapping.get(reference.getReferenceId());
	}
}
