/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/03/09
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.editcommand;

import java.util.List;

import org.apache.commons.collections15.list.UnmodifiableList;
import org.apache.commons.lang.Validate;

import org.jiemamy.EventBroker;
import org.jiemamy.editcommand.Command;
import org.jiemamy.model.ConnectionProfile;
import org.jiemamy.model.DiagramPresentationModel;
import org.jiemamy.model.DiagramPresentations;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.NodeProfile;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.connection.ConnectionAdapter;
import org.jiemamy.model.geometory.JmPoint;
import org.jiemamy.model.geometory.JmRectangle;

/**
 * コネクションのベンドポイントをリセットするEDITコマンド。
 * 
 * <p>基本的にベンドポイントは全て削除されるが、自己コネクションの場合は、見やすさの為のベンドポイントを自動的に2つ設定する。</p>
 * 
 * @author daisuke
 */
public class ResetBendpointCommand extends AbstractCommand {
	
	private static final int AUTO_BENDPOINT_OFFSET = 50;
	
	private final RootModel rootModel;
	
	/** ダイアグラムエディタのインデックス（エディタ内のタブインデックス） */
	private final int diagramIndex;
	
	/** リセット対象のコネクション */
	private final ConnectionAdapter connectionAdapter;
	
	private final DiagramPresentationModel diagramPresentation;
	
	private final ConnectionProfile connectionProfile;
	
	/** リセット前のベンドポイントリスト（undo用） */
	private final List<JmPoint> bendpoints;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param eventBroker イベント通知用{@link EventBroker}
	 * @param rootModel ルートモデル
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter ベンドポイントが削除されるコネクション
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public ResetBendpointCommand(EventBroker eventBroker, RootModel rootModel, int diagramIndex,
			ConnectionAdapter connectionAdapter) {
		super(eventBroker);
		Validate.notNull(rootModel);
		Validate.notNull(connectionAdapter);
		this.rootModel = rootModel;
		this.diagramIndex = diagramIndex;
		this.connectionAdapter = connectionAdapter;
		DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
		diagramPresentation = diagramPresentations.get(diagramIndex);
		connectionProfile = diagramPresentation.getConnectionProfiles().get(connectionAdapter);
		bendpoints = UnmodifiableList.decorate(connectionProfile.getBendpoints());
	}
	
	public Command getNegateCommand() {
		return new RevertBendpointsCommand(getEventBroker(), rootModel, diagramIndex, connectionAdapter, bendpoints);
	}
	
	public JiemamyElement getTarget() {
		return connectionProfile;
	}
	
	@Override
	protected void execute0() {
		List<JmPoint> bendpoints = connectionProfile.getBendpoints();
		bendpoints.clear();
		
		if (connectionAdapter.isSelfConnection()) {
			// 自己コネクション時の、bendpoint自動作成
			
			NodeProfile profile = diagramPresentation.getNodeProfiles().get(connectionAdapter.getSource());
			JmRectangle boundary = profile.getBoundary();
			
			if (boundary == null) {
				throw new IllegalStateException("表示されていないノードのコネクションを操作しようとしました。");
			}
			
			bendpoints.add(0, new JmPoint(Math.max(boundary.x - AUTO_BENDPOINT_OFFSET, 0), boundary.y));
			bendpoints.add(1, new JmPoint(boundary.x, Math.max(boundary.y - AUTO_BENDPOINT_OFFSET, 0)));
		}
	}
}
