/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/08
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.enhance;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.UUID;

import org.apache.commons.lang.Validate;

import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.ReferenceResolveException;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.dialect.generic.GenericDialect;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.experimental.builder.ColumnBuilder;
import org.jiemamy.experimental.builder.TableBuilder;
import org.jiemamy.internal.builder.ColumnBuilderImpl;
import org.jiemamy.internal.builder.TableBuilderImpl;
import org.jiemamy.internal.model.RootModelImpl;
import org.jiemamy.internal.model.attribute.ColumnModelImpl;
import org.jiemamy.internal.model.attribute.ColumnRefImpl;
import org.jiemamy.internal.model.attribute.constraint.ColumnCheckConstraintImpl;
import org.jiemamy.internal.model.attribute.constraint.DeferrabilityImpl;
import org.jiemamy.internal.model.attribute.constraint.ForeignKeyImpl;
import org.jiemamy.internal.model.attribute.constraint.NotNullConstraintImpl;
import org.jiemamy.internal.model.attribute.constraint.PrimaryKeyImpl;
import org.jiemamy.internal.model.attribute.constraint.TableCheckConstraintImpl;
import org.jiemamy.internal.model.attribute.constraint.UniqueKeyImpl;
import org.jiemamy.internal.model.dataset.DataSetModelImpl;
import org.jiemamy.internal.model.dataset.RecordModelImpl;
import org.jiemamy.internal.model.datatype.DomainModelImpl;
import org.jiemamy.internal.model.datatype.DomainRefImpl;
import org.jiemamy.internal.model.entity.TableModelImpl;
import org.jiemamy.internal.model.entity.TableRefImpl;
import org.jiemamy.internal.model.entity.ViewModelImpl;
import org.jiemamy.internal.model.index.IndexColumnModelImpl;
import org.jiemamy.internal.model.index.IndexModelImpl;
import org.jiemamy.model.ElementReference;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.RepresentationAdapter;
import org.jiemamy.model.attribute.constraint.ColumnCheckConstraint;
import org.jiemamy.model.attribute.constraint.Deferrability;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.NotNullConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.TableCheckConstraint;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.datatype.DomainRef;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.model.entity.ViewModel;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.utils.Disablable;
import org.jiemamy.utils.DisablableImpl;

/**
 * factory-enhancerによる拡張ポイントとなる。
 * 
 * @param <T> 生成したいインスタンスの型
 * @author daisuke
 */
public class FactoryExtension<T> {
	
	/**
	 * 新しいアダプタインスタンスを生成する。
	 * 
	 * @param clazz 生成したいインスタンスの型
	 * @return 新しいアダプタインスタンス
	 * @throws IllegalArgumentException 引数に与えたクラスが、生成に対応していない場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	@SuppressWarnings("unchecked")
	public T newAdapter(Class<? extends T> clazz) {
		Validate.notNull(clazz);
		T adapter;
		if (clazz == Disablable.class) {
			adapter = (T) new DisablableImpl();
		} else if (clazz.isInterface() == false) {
			try {
				adapter = clazz.newInstance();
			} catch (InstantiationException e) {
				throw new IllegalArgumentException(e);
			} catch (IllegalAccessException e) {
				throw new IllegalArgumentException(e);
			}
		} else {
			throw new IllegalArgumentException();
		}
		return adapter;
	}
	
	/**
	 * 新しいビルダインスタンスを生成する。
	 * 
	 * @param clazz 生成したいインスタンスの型
	 * @param factory ファクトリ
	 * @return 新しいビルダインスタンス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	@SuppressWarnings("unchecked")
	public T newBuilder(Class<? extends T> clazz, JiemamyFactory factory) {
		Validate.notNull(clazz);
		Validate.notNull(factory);
		T builder = null;
		if (TableBuilder.class.isAssignableFrom(clazz)) {
			builder = (T) new TableBuilderImpl(factory);
		} else if (ColumnBuilder.class.isAssignableFrom(clazz)) {
			builder = (T) new ColumnBuilderImpl(factory);
		}
		return builder;
	}
	
	/**
	 * 新しいインスタンスを生成する。
	 * 
	 * @param clazz 生成したいインスタンスの型
	 * @param jiemamy インスタンス空間
	 * @return 生成されたインスタンス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public T newInstance(Class<? extends T> clazz, Jiemamy jiemamy) {
		Validate.notNull(clazz);
		Validate.notNull(jiemamy);
		return newInstance(clazz, jiemamy, UUID.randomUUID());
	}
	
	/**
	 * 新しいインスタンスを生成する。
	 * 
	 * <p>指定した型が {@link JiemamyElement} のサブクラスであった場合、そのインスタンスは第二引数に指定した
	 * インスタンス空間に属し、第三引数に指定したIDを持つ{@link JiemamyElement}となり、そうでなければ
	 * 第二第三引数は無視される。</p>
	 * 
	 * @param clazz 生成したいインスタンスの型
	 * @param jiemamy インスタンス空間
	 * @param id モデルID
	 * @return 生成されたインスタンス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public T newInstance(Class<? extends T> clazz, Jiemamy jiemamy, UUID id) {
		Validate.notNull(clazz);
		Validate.notNull(jiemamy);
		Validate.notNull(id);
		ModelStrategy modelStrategy = ModelStrategy.fromClass(clazz);
		if (modelStrategy == null) {
			return null;
		}
		return clazz.cast(modelStrategy.newInstance(jiemamy, id));
	}
	
	/**
	 * {@link ColumnModel}から{@link ColumnRef}を生成する。
	 * 
	 * @param columnModel カラム定義
	 * @return カラム参照オブジェクト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public ColumnRef newReference(ColumnModel columnModel) {
		return new ColumnRefImpl(columnModel);
	}
	
	/**
	 * {@link DomainModel}から{@link DomainRef}を生成する。
	 * 
	 * @param domainModel ドメイン定義
	 * @return ドメイン参照オブジェクト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public DomainRef newReference(DomainModel domainModel) {
		return new DomainRefImpl(domainModel);
	}
	
	/**
	 * {@link TableModel}から{@link TableRef}を生成する。
	 * 
	 * @param tableModel テーブル定義
	 * @return テーブル参照オブジェクト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public TableRef newReference(TableModel tableModel) {
		return new TableRefImpl(tableModel);
	}
	

	/**
	 * {@link JiemamyElement}のソートオーダーを決めるコンパレータ。
	 * 
	 * <p>まずFQCNで比較し、同一だったらUUIDで比較。</p>
	 * 
	 * @author daisuke
	 */
	protected enum JiemamyElementComparator implements Comparator<JiemamyElement> {
		
		/** singleton instance */
		INSTANCE;
		
		public int compare(JiemamyElement o1, JiemamyElement o2) {
			if (o1 == o2) {
				return 0;
			}
			if (o1 == null) {
				return -1;
			} else if (o2 == null) {
				return 1;
			}
			int classNameComparison = o1.getClass().getName().compareTo(o2.getClass().getName());
			if (classNameComparison != 0) {
				return classNameComparison;
			}
			
			return o1.getId().compareTo(o2.getId());
		}
	}
	
	/**
	 * 参照オブジェクトのソートオーダーを決めるコンパレータ。
	 * 
	 * @author daisuke
	 */
	protected static class ReferenceComparator implements Comparator<ElementReference<?>> {
		
		private final Jiemamy jiemamy;
		

		ReferenceComparator(Jiemamy jiemamy) {
			this.jiemamy = jiemamy;
			
		}
		
		public int compare(ElementReference<?> o1, ElementReference<?> o2) {
			ReferenceResolver resolver = jiemamy.getReferenceResolver();
			JiemamyElement e1;
			JiemamyElement e2;
			try {
				e1 = resolver.resolve(o1);
				e2 = resolver.resolve(o2);
			} catch (ReferenceResolveException e) {
				e1 = null;
				e2 = null;
			}
			return JiemamyElementComparator.INSTANCE.compare(e1, e2);
		}
	}
	
	/**
	 * モデルインスタンス生成戦略。
	 * 
	 * @author daisuke
	 */
	@SuppressWarnings("unchecked")
	private static enum ModelStrategy {
		/***/
		ROOT(RootModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				RootModelImpl rootModel = new RootModelImpl(jiemamy, id);
				rootModel.setDomains(new ArrayList<DomainModel>());
				rootModel.setEntities(new TreeSet<EntityModel>(JiemamyElementComparator.INSTANCE));
				rootModel.setDataSets(new ArrayList<DataSetModel>());
				rootModel.setDialectClassName(GenericDialect.class.getName());
				return (T) rootModel;
			}
		},
		/***/
		TABLE(TableModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				TableModelImpl tableModel = new TableModelImpl(jiemamy, id);
				tableModel.setAttributes(new ArrayList<AttributeModel>());
				tableModel.setIndexes(new ArrayList<IndexModel>());
				tableModel.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) tableModel;
			}
		},
		/***/
		VIEW(ViewModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				ViewModelImpl viewModel = new ViewModelImpl(jiemamy, id);
				viewModel.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) viewModel;
			}
		},
		/***/
		COLUMN(ColumnModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				ColumnModelImpl columnModel = new ColumnModelImpl(jiemamy, id);
				columnModel.registerAdapter(factory.newAdapter(RepresentationAdapter.class));
				columnModel.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) columnModel;
			}
		},
		/***/
		PK(PrimaryKey.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				PrimaryKeyImpl primaryKey = new PrimaryKeyImpl(jiemamy, id);
				primaryKey.setKeyColumns(new ArrayList<ColumnRef>());
				primaryKey.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) primaryKey;
			}
		},
		/***/
		UNIQ(UniqueKey.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				UniqueKeyImpl uniqueKey = new UniqueKeyImpl(jiemamy, id);
				uniqueKey.setKeyColumns(new ArrayList<ColumnRef>());
				uniqueKey.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) uniqueKey;
			}
		},
		/***/
		FK(ForeignKey.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				ForeignKeyImpl foreignKey = new ForeignKeyImpl(jiemamy, id);
				foreignKey.setKeyColumns(new ArrayList<ColumnRef>());
				foreignKey.setReferenceColumns(new ArrayList<ColumnRef>());
				foreignKey.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) foreignKey;
			}
		},
		/***/
		DEFERRABILITY(Deferrability.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				DeferrabilityImpl deferrability = new DeferrabilityImpl(jiemamy, id);
				return (T) deferrability;
			}
		},
		/***/
		TABLE_CHECK(TableCheckConstraint.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				TableCheckConstraintImpl checkConstraint = new TableCheckConstraintImpl(jiemamy, id);
				checkConstraint.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) checkConstraint;
			}
		},
		/***/
		COLUMN_CHECK(ColumnCheckConstraint.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				ColumnCheckConstraintImpl checkConstraint = new ColumnCheckConstraintImpl(jiemamy, id);
				checkConstraint.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) checkConstraint;
			}
		},
		/***/
		NOT_NULL(NotNullConstraint.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				NotNullConstraintImpl nnModel = new NotNullConstraintImpl(jiemamy, id);
				nnModel.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) nnModel;
			}
		},
		/***/
		DOMAIN(DomainModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				DomainModelImpl domainModel = new DomainModelImpl(jiemamy, id);
				return (T) domainModel;
			}
		},
		/***/
		DATASET(DataSetModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				DataSetModelImpl insertDataSetModel = new DataSetModelImpl(jiemamy, id);
				insertDataSetModel
					.setRecords(new TreeMap<TableRef, List<RecordModel>>(new ReferenceComparator(jiemamy)));
				return (T) insertDataSetModel;
			}
		},
		/***/
		RECORD(RecordModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				RecordModelImpl recordModel = new RecordModelImpl(jiemamy, id);
				recordModel.setValues(new HashMap<ColumnRef, String>());
				return (T) recordModel;
			}
		},
		/***/
		INDEX(IndexModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				JiemamyFactory factory = jiemamy.getFactory();
				IndexModelImpl indexModel = new IndexModelImpl(jiemamy, id);
				indexModel.setColumns(new ArrayList<IndexColumnModel>());
				indexModel.registerAdapter(factory.newAdapter(Disablable.class));
				return (T) indexModel;
			}
		},
		/***/
		INDEX_COLUMN(IndexColumnModel.class) {
			
			@Override
			public <T>T newInstance(Jiemamy jiemamy, UUID id) {
				Validate.notNull(jiemamy);
				Validate.notNull(id);
				IndexColumnModelImpl indexColumnModel = new IndexColumnModelImpl(jiemamy, id);
				return (T) indexColumnModel;
			}
		};
		
		/**
		 * 型からモデルインスタンス生成戦略を取得する。
		 * 
		 * @param clazz 型
		 * @return モデルインスタンス生成戦略. 戦略が存在しない場合は{@code null}
		 */
		public static ModelStrategy fromClass(Class<?> clazz) {
			return ClassToModelStrategy.get(clazz);
		}
		

		/** モデルを特定するためのインターフェイス型 */
		private final Class<?> clazz;
		

		ModelStrategy(Class<? extends JiemamyElement> clazz) {
			assert clazz.isInterface();
			this.clazz = clazz;
		}
		
		/**
		 * モデルインスタンスを生成する。
		 * 
		 * @param <T> 生成するインスタンスの型
		 * @param jiemamy コンテキスト
		 * @param id モデルID
		 * @return モデルインスタンス
		 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
		 */
		public abstract <T>T newInstance(Jiemamy jiemamy, UUID id);
		

		private static final class ClassToModelStrategy {
			
			private static final Map<Class<?>, ModelStrategy> REVERSE_DICTIONARY;
			static {
				Map<Class<?>, ModelStrategy> map = new HashMap<Class<?>, ModelStrategy>();
				for (ModelStrategy elem : ModelStrategy.values()) {
					map.put(elem.clazz, elem);
				}
				REVERSE_DICTIONARY = Collections.unmodifiableMap(map);
			}
			

			static ModelStrategy get(Class<?> key) {
				return REVERSE_DICTIONARY.get(key);
			}
			
			private ClassToModelStrategy() {
				throw new UnexpectedConditionError("不到達ポイント");
			}
		}
		
	}
	
//	@SuppressWarnings("unchecked")
//	private enum ReferenceStrategy {
//		/***/
//		TABLE(TableModel.class) {
//			
//			@Override
//			public <T>T newInstance(JiemamyElement e) {
//				return (T) new TableRefImpl((TableModel) e);
//			}
//		};
//		
//		/**
//		 * 型からモデルインスタンス生成戦略を取得する。
//		 * 
//		 * @param clazz 型
//		 * @return モデルインスタンス生成戦略。戦略が存在しない場合は{@code null}
//		 */
//		public static ReferenceStrategy fromClass(Class<?> clazz) {
//			return ClassToReferenceStrategy.get(clazz);
//		}
//		
//
//		private final Class<?> clazz;
//		
//
//		ReferenceStrategy(Class<?> clazz) {
//			this.clazz = clazz;
//		}
//		
//		/**
//		 * 参照インスタンスを生成する。
//		 * 
//		 * @param <T>
//		 * @param e 
//		 * @return 参照インスタンス
//		 */
//		public abstract <T>T newInstance(JiemamyElement e);
//		
//
//		private static final class ClassToReferenceStrategy {
//			
//			private static final Map<Class<?>, ReferenceStrategy> REVERSE_DICTIONARY;
//			static {
//				Map<Class<?>, ReferenceStrategy> map = new HashMap<Class<?>, ReferenceStrategy>();
//				for (ReferenceStrategy elem : ReferenceStrategy.values()) {
//					map.put(elem.clazz, elem);
//				}
//				REVERSE_DICTIONARY = Collections.unmodifiableMap(map);
//			}
//			
//
//			static ReferenceStrategy get(Class<?> key) {
//				return REVERSE_DICTIONARY.get(key);
//			}
//			
//			private ClassToReferenceStrategy() {
//			}
//		}
//	}
}
