/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/09/17
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.model;

import java.util.Collection;
import java.util.List;
import java.util.SortedSet;

import org.jiemamy.dialect.Dialect;
import org.jiemamy.exception.ElementNotFoundException;
import org.jiemamy.exception.TooManyElementsException;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;

/**
 * Jiemamyの根となるモデルを表すモデルインターフェイス。
 * 
 * <p>この型を持つインスタンス1つにつき、1つのDB状態を表現する。</p>
 * 
 * @since 0.2
 * @author daisuke
 */
public interface RootModel extends JiemamyElement {
	
	/**
	 * 開始スクリプトを取得する。
	 * 
	 * @return 開始スクリプト. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	String getBeginScript();
	
	/**
	 * INSERTデータセットのリストを取得する。
	 * 
	 * <p>このメソッドは、インスタンスの持つフィールドをそのまま返す。返される{@link List}を直接操作することで、
	 * このオブジェクトのフィールドとして保持される{@link List}を変更することができる。</p>
	 * 
	 * @return INSERTデータセットのリスト
	 * @since 0.2
	 */
	List<DataSetModel> getDataSets();
	
	/**
	 * 説明文を取得する。
	 * 
	 * @return 説明文. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	String getDescription();
	
	/**
	 * SQL方言を取得する。
	 * 
	 * @return SQL方言
	 * @throws ClassNotFoundException 設定されたクラスが見つからなかった場合 
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws UnsupportedOperationException 実装がこのメソッドをサポートしない場合
	 */
	Dialect findDialect() throws ClassNotFoundException;
	
	/**
	 * SQL方言IDを取得する。
	 * 
	 * @return SQL方言クラスのFQCN. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	String getDialectClassName();
	
	/**
	 * ドメイン定義のリストを取得する。
	 * 
	 * <p>このメソッドは、インスタンスの持つフィールドをそのまま返す。返される{@link List}を直接操作することで、
	 * このオブジェクトのフィールドとして保持される{@link List}を変更することができる。</p>
	 * 
	 * @return ドメイン定義のリスト
	 * @since 0.2
	 */
	List<DomainModel> getDomains();
	
	/**
	 * 終了スクリプトを取得する。
	 * 
	 * @return 終了スクリプト. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	String getEndScript();
	
	/**
	 * エンティティ定義のセットを取得する。
	 * 
	 * <p>このメソッドは、インスタンスの持つフィールドをそのまま返す。返される{@link SortedSet}を直接操作することで、
	 * このオブジェクトのフィールドとして保持される{@link SortedSet}を変更することができる。</p>
	 * 
	 * <p>このメソッドが返すSortedSetは、まずエンティティ実装クラスのFQCN（完全修飾クラス名）で昇順ソートされ、
	 * FQCNが同一の場合、モデルIDで昇順ソートされる。</p>
	 * 
	 * @return エンティティ定義のセット
	 * @since 0.2
	 */
	SortedSet<EntityModel> getEntities();
	
	/**
	 * 特定の種類のエンティティのリストを取得する。
	 * 
	 * <p>返される{@link Collection}は他に影響を及ぼさない独立したインスタンスである。</p>
	 * 
	 * @param <T> 取得したいエンティティのクラス
	 * @param clazz 取得したいエンティティのクラス
	 * @return ノードのリスト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws UnsupportedOperationException 実装がこのメソッドをサポートしない場合
	 */
	<T extends EntityModel>Collection<T> findEntities(Class<T> clazz);
	
	/**
	 * 指定した名前を持つエンティティを取得する。
	 * 
	 * <p>エンティティ名の大文字・小文字は区別されない。</p>
	 * 
	 * @param <T> エンティティの種類を表すクラス
	 * @param clazz エンティティの種類を表すクラス. {@code null}の場合、エンティティの種類を限定しないことを表す
	 * @param entityName エンティティ名
	 * @return 見つかったエンティティ
	 * @throws TooManyElementsException 同名のエンティティが複数見つかった場合
	 * @throws ElementNotFoundException エンティティが見つからなかった場合
	 * @throws IllegalArgumentException 引数rootModel, entityNameに{@code null}を与えた場合
	 * @throws UnsupportedOperationException 実装がこのメソッドをサポートしない場合
	 */
	<T extends EntityModel>T getEntity(Class<T> clazz, String entityName);
	
	/**
	 * スキーマ名を取得する。
	 * 
	 * @return スキーマ名. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	String getSchemaName();
	
	/**
	 * モデルの正規化を行う。
	 * 
	 * TODO もっと説明
	 * 
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws UnsupportedOperationException 実装がこのメソッドをサポートしない場合
	 */
	void normalize();
	
	/**
	 * 開始スクリプトを設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param beginScript 開始スクリプト
	 * @since 0.2
	 */
	void setBeginScript(String beginScript);
	
	/**
	 * 説明文を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param description 説明文
	 * @since 0.2
	 */
	void setDescription(String description);
	
	/**
	 * SQL方言を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param dialectClassName SQL方言クラス名
	 * @since 0.2
	 */
	void setDialectClassName(String dialectClassName);
	
	/**
	 * 終了スクリプトを設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param endScript 終了スクリプト
	 * @since 0.2
	 */
	void setEndScript(String endScript);
	
	/**
	 * スキーマ名を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param schemaName スキーマ名
	 * @since 0.2
	 */
	void setSchemaName(String schemaName);
	
}
