/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/07/19
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils;

import java.util.List;

import org.jiemamy.model.Adapter;

/**
 * アダプタを適用可能な型である事を示すインターフェイス。
 * 
 * <p>Extension Objectパターンを実現するインターフェイス。
 * Eclipseに実装されたIAdaptableインターフェイスと同様の考え方のインターフェイスである。</p>
 * 
 * <p>cf. http://d.hatena.ne.jp/daisuke-m/20081212/1229083116</p>
 * 
 * @since 0.2
 * @author daisuke
 */
public interface Adaptable {
	
	/**
	 * このオブジェクトにまつわる、指定した型を持つアダプタのインスタンスを取得する。
	 * 
	 * <p>このメソッドは{@code null}を返さない。</p>
	 * 
	 * 指定した型に一致するものが複数あった場合は、より先に登録されたアダプタが返される。
	 * 
	 * @param <T> アダプタを表すクラス
	 * @param adapterClass アダプタを表すクラス
	 * @return アダプタのインスタンス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws AdapterNotFoundException 登録されていないアダプタを取得しようとした場合
	 * @since 0.2
	 */
	<T>T getAdapter(Class<T> adapterClass);
	
	/**
	 * このオブジェクトが持つアダプタのリストを取得する。
	 * 
	 * <p>このメソッドは、インスタンスの持つ{@link List}の防御コピーを返す。返される{@link List}を直接操作しても、
	 * このオブジェクトの属性を操作することはできない。ただし、コレクションの要素はコピーされない、シャローコピーである。</p>
	 * 
	 * @return アダプタのリスト
	 * @throws UnsupportedOperationException アダプタのリストを提供しない場合
	 * @since 0.2
	 */
	List<Object> getAdapters();
	
	/**
	 * 指定したアダプタを持つ（取得可能）かどうか調べる。
	 * 
	 * @param adapterClass アダプタクラス
	 * @return 適用可能の場合{@code true}、そうでない場合は{@code false}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	boolean hasAdapter(Class<?> adapterClass);
	
	/**
	 * アダプタを登録する。
	 * 
	 * <p>アダプタは、その実装クラスまたは上流の型のいずれかに{@link Adapter}アノテーションをつけなければならない。
	 * また、その場合そのアダプタは、同一性判別のため{@link Object#equals(Object)}メソッドをオーバーライド
	 * していなければならない。</p>
	 * 
	 * @param adapter 登録するアダプタ
	 * @throws UnsupportedOperationException アダプタの追加登録をサポートしない場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws IllegalArgumentException 引数のクラスに{@link Adapter}アノテーションがついていない場合
	 * @since 0.2
	 */
	void registerAdapter(Object adapter);
	
	/**
	 * 指定した型のアダプタを削除する。
	 * 
	 * <p>{@link #getAdapter(Class)}で取得されるアダプタが削除される。</p>
	 * 
	 * @param <T> アダプタを表すクラス
	 * @param adapterClass アダプタクラス
	 * @return 削除されたアダプタのインスタンス
	 * @since 0.2
	 */
	<T>T unregisterAdapter(Class<T> adapterClass);
}
