/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/02/06
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils;

import java.util.Comparator;
import java.util.List;

import org.apache.commons.collections15.list.UnmodifiableList;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.Validate;

import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.TableCheckConstraint;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.entity.TableModel;

/**
 * 属性の出力順を整列させるコンパレータ。
 * 
 * <p>カラム, PK, UK, FK, CHECKの順にならべ、同種内では元の順序を維持する。</p>
 * 
 * @author daisuke
 */
public class AttributeComparator implements Comparator<AttributeModel> {
	
	static final List<Class<? extends AttributeModel>> ORDER;
	
	static {
		List<Class<? extends AttributeModel>> order = CollectionsUtil.newArrayList();
		order.add(ColumnModel.class);
		order.add(PrimaryKey.class);
		order.add(UniqueKey.class);
		order.add(ForeignKey.class);
		order.add(TableCheckConstraint.class);
		ORDER = UnmodifiableList.decorate(order);
	}
	

	private static int getOrder(Class<?>[] interfaces) {
		for (Class<?> interf : interfaces) {
			int index = ORDER.indexOf(interf);
			if (index != ArrayUtils.INDEX_NOT_FOUND) {
				return index;
			}
		}
		return -1;
	}
	

	private final List<AttributeModel> originalOrder;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param tableModel ソート対象の属性が保持されているテーブル
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public AttributeComparator(TableModel tableModel) {
		Validate.notNull(tableModel);
		originalOrder = CollectionsUtil.newArrayList(tableModel.getAttributes());
	}
	
	public int compare(AttributeModel o1, AttributeModel o2) {
		if (o1 == o2) {
			return 0;
		}
		if (o1 == null) {
			return -1;
		} else if (o2 == null) {
			return 1;
		}
		int i1 = getOrder(o1.getClass().getInterfaces());
		int i2 = getOrder(o2.getClass().getInterfaces());
		if (i1 != i2) {
			return i1 - i2;
		}
		
		return originalOrder.indexOf(o1) - originalOrder.indexOf(o2);
	}
}
