/*
 * Copyright 2009 Jiemamy Project and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.enhancer.helper;

import java.lang.reflect.Modifier;
import java.text.MessageFormat;
import java.util.regex.Pattern;

import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMember;
import javassist.CtMethod;
import javassist.NotFoundException;
import javassist.expr.NewExpr;

import org.jiemamy.utils.enhancer.EnhanceException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * エンハンスを行う際の定型的なクラス書き換えを行うためのライブラリ。
 * @version $Date: 2009-01-21 20:37:40 +0900 (水, 21  1月 2009) $
 * @author Suguru ARAKAWA (Gluegent, Inc.)
 */
public class EnhanceManipulator {
    
    private static final Logger LOG =
        LoggerFactory.getLogger(EnhanceManipulator.class);

    private static final Pattern AVOID_NAME_PATTERN = Pattern.compile(
        "(java\\.|javax\\.|sun\\.).*"); //$NON-NLS-1$
    
    private static final String ENHANCE_CLASS = "__ENHANCED__";

    private static final String BYPASS_METHOD = "__BYPASS__"; //$NON-NLS-1$

    private static final String ADVICE_TABLE_FIELD = "__ADVICE_TABLE__"; //$NON-NLS-1$

    /**
     * 指定のクラスをコピーした新しいクラスを作成して返す。
     * <p>
     * 作成したクラスは元のクラスの名称を変更したクラスで、すべての内容は元のクラスと同一である。
     * </p>
     * @param klass コピー元のクラス
     * @return 作成したクラス
     * @throws EnhanceException クラスのコピーに失敗した場合
     */
    public static CtClass createCopyClass(CtClass klass) throws EnhanceException {
        assert klass != null;
        assert klass.isInterface() == false && klass.isEnum() == false;
        LOG.trace("Creating a copy: {}", klass);
        String name = klass.getName();
        try {
            CtClass copy = klass.getClassPool().getAndRename(
                name, getEnhanceClassName(name));
            copy.setModifiers(klass.getModifiers() | Modifier.FINAL);
            LOG.debug("Copy class: {} -> {}", klass.getName(), copy.getName());
            return copy;
        }
        catch (NotFoundException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot load {0}",
                    name),
                e);
        }
    }

    /**
     * 指定のクラスを継承した新しいクラスを作成して返す。
     * <p>
     * 生成されたクラスは、引数にとるクラスを親クラスに持ち、
     * 親クラスで宣言されたコンストラクタのうちジョインポイントとして利用可能なコンストラクタに対する
     * 移譲コンストラクタを宣言する。
     * </p>
     * @param klass 元となるクラス
     * @return 対象のクラスを継承した新しいクラス
     * @throws EnhanceException クラスの作成に失敗した場合
     * @see #createDelegateConstructors(CtClass)
     */
    public static CtClass createInheritedClass(CtClass klass)
            throws EnhanceException {
        assert klass != null;
        assert klass.isInterface() == false && klass.isEnum() == false;
        LOG.trace("Creating an inherited class: {}", klass.getName());
        CtClass copy = klass.getClassPool().makeClass(
            getEnhanceClassName(klass.getName()));
        copy.setModifiers(Modifier.PUBLIC | Modifier.FINAL);
        try {
            copy.setSuperclass(klass);
        }
        catch (CannotCompileException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot inherit {0}",
                    klass.getName()),
                e);
        }
        LOG.debug("Inherit class: {} -> {}", klass.getName(), copy.getName());
        createDelegateConstructors(copy);
        return copy;
    }

    /**
     * 親クラスで宣言されたコンストラクタに処理を委譲するコンストラクタを拡張クラス上に宣言する。
     * <p>
     * 拡張クラス上に宣言される移譲コンストラクタは、移譲先のコンストラクタと同一の引数リストを持ち、
     * かつ与えられた引数リストをそのまま移譲先に渡すようなコンストラクタである。
     * </p>
     * <p>
     * この呼び出しによって作成される移譲コンストラクタは、
     * 移譲先のコンストラクタがジョインポイントとして適切である必要がある。
     * ジョインポイントとして適切でないコンストラクタについては、拡張クラス上に移譲コンストラクタを作成しない。
     * </p>
     * @param enhanced 宣言する対象のクラス
     * @throws EnhanceException コンストラクタの宣言に失敗した場合
     */
    private static void createDelegateConstructors(CtClass enhanced)
            throws EnhanceException {
        assert enhanced != null;
        LOG.trace("Creating delegate constructors: {}", enhanced.getName());
        CtClass base;
        try {
            base = enhanced.getSuperclass();
        }
        catch (NotFoundException e) {
            // may not occur
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot resolve superclass for {0}",
                    enhanced.getName()),
                e);
        }
        for (CtConstructor constructor : base.getConstructors()) {
            if (Modifier.isPublic(constructor.getModifiers()) == false) {
                LOG.debug("Skip delegate constructor: {}{}",
                    base.getName(),
                    constructor.getSignature());
                continue;
            }
            if (constructor.getName().charAt(0) == '<') {
                continue;
            }
            try {
                CtConstructor delegate =
                    new CtConstructor(constructor.getParameterTypes(), enhanced);
                delegate.setExceptionTypes(constructor.getExceptionTypes());
                delegate.setModifiers(Modifier.PUBLIC);
                delegate.setBody("super($$);");
                enhanced.addConstructor(delegate);
    
                LOG.debug("Delegate constructor: {}{}",
                    enhanced.getName(),
                    delegate.getSignature());
            }
            catch (NotFoundException e) {
                throw new EnhanceException(
                    MessageFormat.format(
                        "Cannot inspect constructor ({0})",
                        constructor.getLongName()),
                    e);
            }
            catch (CannotCompileException e) {
                throw new EnhanceException(
                    MessageFormat.format(
                        "Cannot delegate constructor ({0})",
                        constructor.getLongName()),
                    e);
            }
        }
    }

    /**
     * {@code method}に指定されたメソッドを呼び出すバイパスメソッドを拡張クラス上に新しく作成して返す。
     * <p>
     * バイパスメソッドは、拡張元のクラスが公開するメソッドを安全に呼び出すための方法を提供する。
     * 親クラスで定義されたメソッドは、アドバイスを実行するためのメソッドで上書きしてしまうため
     * 上書きされたメソッドを迂回して元のメソッドを呼び出す別のメソッドを定義している。
     * </p>
     * <pre><code>
     * // original class
     * public class Hoge {
     *   public int add(int a, int b) {
     *     return a + b;
     *   }
     * }
     * </code></pre>
     * <p>
     * 上記のクラスは、拡張された際に下記のようなバイパスメソッドを追加される。
     * ただし、バイパスメソッドの名前は他のメソッドの名称と衝突しないようなものが自動的に選択される。
     * </p>
     * <pre><code>
     * // enhanced class (extends original class)
     * public class Hoge__Enhanced__ extends Hoge {
     *   ...
     *   // 上書きしてアスペクトを処理する
     *   &#64;Override public int add(int a, int b) {
     *     return (invoke advice chain);
     *   }
     *   // 親クラス(Hoge)のaddメソッドを呼び出すためのバイパス
     *   public int bypass(int a, int b) {
     *     return super.add(int a, int b);
     *   }
     * }
     * </code></pre>
     * <p>
     * なお、このメソッドでは拡張するクラスにバイパスメソッドを追加するだけで、
     * アスペクトを処理するメソッドについては
     * {@link #createPointcutMethod(CtClass, CtMethod, CtField, int)}
     * で作成する。
     * </p>
     * @param enhance 拡張対象のクラス
     * @param method バイパスメソッドが呼び出す対象のメソッド
     * @param index 拡張メソッドの番号
     * @return 作成したバイパスメソッド
     * @throws EnhanceException バイパスメソッドの作成に失敗した場合
     * @see #createPointcutMethod(CtClass, CtMethod, CtField, int)
     */
    public static CtMethod createBypassMethod(CtClass enhance,
            CtMethod method, int index) throws EnhanceException {
        
        assert enhance != null;
        assert method != null;
        assert index >= 0;
        assert enhance.subclassOf(method.getDeclaringClass());
        assert isStatic(method) == false;
        
        LOG.trace("Creating bypass method: {}#{}",
            method.getDeclaringClass().getName(), method.getName());
        
        try {
            CtMethod bypass = new CtMethod(
                method.getReturnType(),
                getBypassMethodName(method, index),
                method.getParameterTypes(),
                enhance);
            
            LOG.debug("Bypass method: {}#{}{} -> {}#{}{}", new Object[] {
                    bypass.getDeclaringClass().getName(),
                    bypass.getName(),
                    bypass.getSignature(),
                    method.getDeclaringClass().getName(),
                    method.getName(),
                    method.getSignature()
            });
            bypass.setModifiers(Modifier.PUBLIC);
            if (method.getReturnType() == CtClass.voidType) {
                bypass.setBody(String.format(
                    "super.%s($$);",
                    method.getName()));
            }
            else {
                bypass.setBody(String.format(
                    "return super.%s($$);",
                    method.getName()));
            }
            enhance.addMethod(bypass);
            return bypass;
        }
        catch (NotFoundException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot create bypass method for {0}",
                    method.getLongName()),
                e);
        }
        catch (CannotCompileException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot create bypass method for {0}",
                    method.getLongName()),
                e);
        }
    }

    /**
     * {@code method}の動作をフックするポイントカットメソッドを拡張クラス上に作成して返す。
     * <p>
     * ポイントカットメソッドは、拡張元のクラスが公開するメソッドを上書きして
     * 登録されたアドバイスを実行するための機能を提供する。
     * ポイントカットメソッドが実行するアドバイスは、
     * {@link #createAdviceTableField(CtClass)}によって作成されるフィールドに含まれている。
     * </p>
     * @param enhance 拡張対象のクラス
     * @param method フックする対象のメソッド
     * @param holder アドバイスを保持するフィールド
     * @param index 拡張メソッドの番号
     * @return 作成したポイントカットメソッド
     * @throws EnhanceException ポイントカットメソッドの作成に失敗した場合
     * @see #createAdviceTableField(CtClass)
     * @see #createPointcutMethod(CtClass, CtMethod, CtField, int)
     */
    public static CtMethod createPointcutMethod(CtClass enhance,
            CtMethod method, CtField holder, int index)
            throws EnhanceException {
        
        assert enhance != null;
        assert method != null;
        assert enhance.subclassOf(method.getDeclaringClass());
        assert holder != null;
        assert enhance.equals(holder.getDeclaringClass());
        assert index >= 0;
        assert isStatic(method) == false;
        
        LOG.trace("Creating pointcut method: {}#{}",
            method.getDeclaringClass().getName(), method.getName());

        try {
            CtMethod pointcut = new CtMethod(
                method.getReturnType(),
                method.getName(),
                method.getParameterTypes(),
                enhance);

            LOG.debug("Pointcut method[{}]: {}#{}{}", new Object[] {
                    index,
                    pointcut.getDeclaringClass().getName(),
                    pointcut.getName(),
                    pointcut.getSignature()
            });
            
            pointcut.setModifiers(Modifier.PUBLIC);
            if (method.getReturnType() == CtClass.voidType) {
                pointcut.setBody(String.format(
                    "%s[%d].invoke(this, $args);",
                    holder.getName(),
                    index));
            }
            else {
                pointcut.setBody(String.format(
                    "return ($r) %s[%d].invoke(this, $args);",
                    holder.getName(),
                    index));
            }
            enhance.addMethod(pointcut);
            return pointcut;
        }
        catch (NotFoundException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot create pointcut method for {0}",
                    method.getLongName()),
                e);
        }
        catch (CannotCompileException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot create pointcut method for {0}",
                    method.getLongName()),
                e);
        }
    }

    /**
     * ポイントカットメソッドが参照するアドバイステーブルを保持するフィールド拡張クラス上に作成して返す。
     * <p>
     * アドバイステーブルは、それぞれのポイントカットメソッドが利用するアドバイスを提供する。
     * このメソッドによって作成されるフィールドは、このアドバイステーブルを保持するために
     * それぞれのクラス上に自動的に宣言される。
     * それぞれのアドバイスは{@link AdviceApplier}クラスによって定義され、
     * フィールドにはこの配列型の値が格納される。
     * </p>
     * <p>
     * それぞれのポイントカットメソッドには「拡張メソッド番号」というものが{@code 0}から順に
     * 一意に割り当てられ、その番号と同一のインデックスを利用してこのフィールドの
     * 配列に含まれるアドバイスを参照する。
     * </p>
     * @param target 拡張対象のクラス
     * @return 作成した
     * @throws EnhanceException フィールドの作成に失敗した場合
     * @see #createPointcutMethod(CtClass, CtMethod, CtField, int)
     */
    public static CtField createAdviceTableField(CtClass target)
            throws EnhanceException {
        
        assert target != null;
        LOG.trace("Creating advice table field: {}", target.getName());
        try {
            CtField field  = new CtField(
                getAdviceTableType(target),
                getAdviceTableFieldName(target),
                target);
            field.setModifiers(Modifier.PUBLIC | Modifier.STATIC);
            target.addField(field);
            LOG.debug("Advice table: {}#{}",
                field.getDeclaringClass().getName(),
                field.getName());
            return field;
        }
        catch (NotFoundException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot create advice table field for {0}",
                    target.getName()),
                e);
        }
        catch (CannotCompileException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot create advice table field for {0}",
                    target.getName()),
                e);
        }
    }
    
    /**
     * 指定のクラス名に対する拡張クラスの名称を返す。
     * @param name 対象のクラス名
     * @return 拡張クラスの名称
     */
    private static String getEnhanceClassName(String name) {
        assert name != null;
        if (AVOID_NAME_PATTERN.matcher(name).matches()) {
            return ENHANCE_CLASS + name;
        }
        return name + ENHANCE_CLASS;
    }

    /**
     * 指定のメソッドに対するバイパスメソッドの名称を返す。
     * @param method 対象のメソッド
     * @param index 拡張メソッド番号
     * @return バイパスメソッドの名称
     */
    private static String getBypassMethodName(CtMethod method, int index) {
        assert method != null;
        assert index >= 0;
        return String.format(
            "%08d%s",
            index,
            BYPASS_METHOD);
    }

    /**
     * 指定のクラスに対するアドバイステーブルフィールドの名称を返す。
     * @param target 対象のクラス
     * @return アドバイステーブルフィールドの名称
     */
    private static String getAdviceTableFieldName(CtClass target) {
        assert target != null;
        return ADVICE_TABLE_FIELD;
    }

    /**
     * アドバイステーブルの型を計算して返す。
     * @param klass 対象の型
     * @return アドバイステーブルの型
     * @throws NotFoundException アドバイステーブルの型をロードできなかった場合
     */
    private static CtClass getAdviceTableType(CtClass klass) throws NotFoundException {
        assert klass != null;
        CtClass helper = klass.getClassPool().get(
            AdviceApplier.class.getName() + "[]"); //$NON-NLS-1$
        return helper;
    }
    
    /**
     * {@code target}の動作をフックするポイントカット実行式を、
     * 元の{@code target}の処理の代わりに上書きする。
     * <p>
     * ポイントカット実行式は、拡張元のクラスが実行するインスタンス生成を上書きして、
     * 登録されたアドバイスを実行するための機能を提供する。
     * ポイントカットメソッドが実行するアドバイスは、
     * {@link #createAdviceTableField(CtClass)}によって作成されるフィールドに含まれている。
     * </p>
     * @param target フックする対象のインスタンス生成式
     * @param holder アドバイスを保持するフィールド
     * @param index 拡張番号
     * @throws CannotCompileException コンパイルに失敗した場合
     * @throws NullPointerException 引数に{@code null}が指定された場合
     */
    public static void replaceToPointcut(NewExpr target, CtField holder,
            int index) throws CannotCompileException {
        
        if (target == null) {
            throw new NullPointerException("expr");
        }
        
        CtBehavior contextBehaviour = target.where();
        if (isStatic(contextBehaviour)) {
            target.replace(String.format(
                "$_ = ($r) %s[%d].invoke(%s, $args);", //$NON-NLS-1$
                holder.getName(),
                index,
                contextBehaviour.getDeclaringClass().getName() + ".class")); //$NON-NLS-1$
        }
        else {
            target.replace(String.format(
                "$_ = ($r) %s[%d].invoke(this, $args);", //$NON-NLS-1$
                holder.getName(),
                index));
        }
    }
    
    /**
     * 指定のメンバが{@code static}として宣言されている場合のみ{@code true}を返す。
     * @param member 対象のメンバ
     * @return
     *      {@code static}として宣言されている場合に{@code true}、
     *      そうでない場合は{@code false}
     */
    private static boolean isStatic(CtMember member) {
        assert member != null;
        return Modifier.isStatic(member.getModifiers());
    }

    private EnhanceManipulator() {
        super();
    }
}
