/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/22
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.model;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.List;

import au.com.bytecode.opencsv.CSVReader;
import au.com.bytecode.opencsv.CSVWriter;

import org.apache.commons.lang.Validate;

import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.utils.CollectionsUtil;

/**
 * {@link DataSetModel}のユーティリティクラス。
 * 
 * @author daisuke
 */
public final class DataSetUtil {
	
	/**
	 * 1テーブル分のレコードデータをすべてストリームにエクスポートする。
	 * 
	 * @param dataSetModel データセット
	 * @param tableModel 対象テーブル
	 * @param out 出力ストリーム
	 * @throws IOException 入出力エラーが発生した場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static void exportToCsv(DataSetModel dataSetModel, TableModel tableModel, OutputStream out)
			throws IOException {
		Validate.notNull(dataSetModel);
		Validate.notNull(tableModel);
		Validate.notNull(out);
		Jiemamy jiemamy = tableModel.getJiemamy();
		JiemamyFactory factory = jiemamy.getFactory();
		
		TableRef tableRef = factory.newReference(tableModel);
		List<RecordModel> records = dataSetModel.getRecords().get(tableRef);
		
		CSVWriter writer = null;
		try {
			writer = new CSVWriter(new OutputStreamWriter(out));
			
			List<ColumnModel> columns = tableModel.findColumns();
			String[] line = new String[columns.size()];
			int i = 0;
			for (ColumnModel columnModel : columns) {
				line[i++] = columnModel.getName();
			}
			writer.writeNext(line);
			
			for (RecordModel recordModel : records) {
				line = new String[columns.size()];
				for (ColumnModel columnModel : columns) {
					int index = columns.indexOf(columnModel);
					ColumnRef columnRef = factory.newReference(columnModel);
					line[index] = recordModel.getValues().get(columnRef);
				}
				writer.writeNext(line);
			}
		} finally {
			if (writer != null) {
				writer.close();
			}
		}
	}
	
	/**
	 * 1テーブル分のレコードデータをすべてストリームからインポートする。
	 * 
	 * <p>既にセットされているレコードはすべてクリアされる。</p>
	 * 
	 * @param dataSetModel データセット
	 * @param tableModel 対象テーブル
	 * @param in 入力ストリーム
	 * @throws IOException 入出力エラーが発生した場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static void importFromCsv(DataSetModel dataSetModel, TableModel tableModel, InputStream in)
			throws IOException {
		Validate.notNull(dataSetModel);
		Validate.notNull(tableModel);
		Validate.notNull(in);
		
		JiemamyFactory factory = tableModel.getJiemamy().getFactory();
		TableRef tableRef = factory.newReference(tableModel);
		
		List<RecordModel> records = dataSetModel.getRecords().get(tableRef);
		records.clear();
		
		CSVReader reader = null;
		try {
			reader = new CSVReader(new InputStreamReader(in));
			String[] line = reader.readNext();
			
			List<ColumnModel> columnModels = CollectionsUtil.newArrayList();
			for (String columnName : line) {
				ColumnModel columnModel = tableModel.findColumn(columnName);
				columnModels.add(columnModel);
			}
			
			while ((line = reader.readNext()) != null) {
				RecordModel recordModel = factory.newModel(RecordModel.class);
				for (int i = 0; i < line.length; i++) {
					ColumnRef columnRef = factory.newReference(columnModels.get(i));
					recordModel.getValues().put(columnRef, line[i]);
				}
				records.add(recordModel);
			}
		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}
	
	/**
	 * {@link DataSetModel}にレコードを設定する。
	 * 
	 * @param dataSetModel 対象{@link DataSetModel}
	 * @param tableModel 設定対象テーブル
	 * @param records レコード
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static void setRecord(DataSetModel dataSetModel, TableModel tableModel, List<RecordModel> records) {
		Validate.notNull(dataSetModel);
		Validate.notNull(tableModel);
		Validate.notNull(records);
		JiemamyFactory factory = tableModel.getJiemamy().getFactory();
		TableRef ref = factory.newReference(tableModel);
		dataSetModel.getRecords().put(ref, records);
	}
	
	private DataSetUtil() {
		throw new UnexpectedConditionError("不到達ポイント");
	}
}
