/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/30
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.model;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.jiemamy.Jiemamy;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.model.Adapter;
import org.jiemamy.model.AdapterType;
import org.jiemamy.model.ElementReference;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.datatype.DomainRef;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.utils.AnnotationUtil;
import org.jiemamy.utils.gtree.converter.CachedObjectConverter;
import org.jiemamy.utils.gtree.converter.ConverterDriver;
import org.jiemamy.utils.gtree.converter.InterfaceBeanConverter;
import org.jiemamy.utils.gtree.converter.ObjectConverter;
import org.jiemamy.utils.gtree.model.Entry;
import org.jiemamy.utils.gtree.model.Record;
import org.jiemamy.utils.gtree.model.Terminal;
import org.jiemamy.utils.gtree.model.Value;
import org.jiemamy.utils.gtree.text.Emitter;

/**
 * 2つのオブジェクトの比較を行うユーティリティクラス。
 * 
 * @author daisuke
 */
public final class EqualsUtil {
	
	/** for debug */
	public static String leftGTree;
	
	/** for debug */
	public static String rightGTree;
	

	/**
	 * 2つのオブジェクトのプロパティが等しいかどうか調べる。
	 * 
	 * @param left 比較左辺
	 * @param right 比較右辺
	 * @return 等しい場合は{@code true}、そうでない場合は{@code false}
	 */
	public static boolean equals(Object left, Object right) {
		leftGTree = null;
		rightGTree = null;
		if (left == right) {
			return true;
		}
		if (left == null || right == null) {
			return false;
		}
		if (left.getClass() != right.getClass()) {
			return false;
		}
		
		// コンバータを作る
		ObjectConverter converter =
				CachedObjectConverter.newInstance(Arrays.asList(new JiemamyPropertyConverter(), new AdapterConverter(),
						new ReferenceConverter(), new InterfaceBeanConverter()));
		
		// コンバータでモデルをGeneric Treeに変換
		Value leftGenericTree = converter.convert(left);
		Value rightGenericTree = converter.convert(right);
		
		// エミッタでPrintWriterに出力
		StringWriter leftOut = new StringWriter();
		StringWriter rightOut = new StringWriter();
		PrintWriter leftWriter = new PrintWriter(leftOut);
		PrintWriter rightWriter = new PrintWriter(rightOut);
		Emitter.emit(leftGenericTree, leftWriter);
		Emitter.emit(rightGenericTree, rightWriter);
		leftWriter.close();
		rightWriter.close();
		leftGTree = leftOut.toString();
		rightGTree = rightOut.toString();
		
		return leftGTree.equals(rightGTree);
	}
	
	private EqualsUtil() {
		throw new UnexpectedConditionError("不到達ポイント");
	}
	

	private static class AdapterConverter implements ConverterDriver {
		
		public Value convert(Object object, ObjectConverter converter) {
			Adapter annotation = AnnotationUtil.getTypeAnnotation(object, Adapter.class);
			if (annotation == null) {
				return null;
			}
			if (annotation.value() == AdapterType.LOGIC) {
				return Terminal.of("<LogicAdapter>");
			} else if (annotation.value() == AdapterType.MODEL) {
				List<Entry> entries = new ArrayList<Entry>();
				for (Field f : object.getClass().getDeclaredFields()) {
					f.setAccessible(true);
					try {
						Object value = f.get(object);
						Entry entry = Entry.of(Terminal.of(f.getName()), converter.convert(value));
						entries.add(entry);
					} catch (IllegalArgumentException e) {
						throw new UnexpectedConditionError("フィールドは存在するはず。");
					} catch (IllegalAccessException e) {
						throw new UnexpectedConditionError("setAccessible(true)したはず。");
					}
				}
				
				return Record.of(entries);
			}
			
			return null;
		}
		
	}
	
	private static class JiemamyPropertyConverter implements ConverterDriver {
		
		public Value convert(Object object, ObjectConverter converter) {
			if (object instanceof Jiemamy) {
				return Terminal.of("<JIEMAMY>");
			}
			return null;
		}
		
	}
	
	private static class ReferenceConverter implements ConverterDriver {
		
		public Value convert(Object object, ObjectConverter converter) {
			if (object instanceof ElementReference) {
				ElementReference<?> ref = (ElementReference<?>) object;
				
				String key = "Ref";
				if (ref instanceof TableRef) {
					key = "TableRef";
				} else if (ref instanceof ColumnRef) {
					key = "ColumnRef";
				} else if (ref instanceof DomainRef) {
					key = "DomainRef";
				}
				
				return Terminal.of(key + "(" + ref.getReferenceId().toString() + ")");
			}
			return null;
		}
		
	}
}
