/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.model;

import java.util.Iterator;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;

import org.jiemamy.ReferenceResolver;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.utils.CollectionsUtil;

/**
 * {@link KeyConstraint}のユーティリティクラス。
 * 
 * @author daisuke
 */
public class KeyConstraintUtil {
	
	/**
	 * {@link ColumnModel}を {@link ColumnRef}に変換し、キーカラムに追加する。
	 * 
	 * @param keyModel 対象キー制約モデル
	 * @param columnModel 追加するカラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static void addKeyColumn(KeyConstraint keyModel, ColumnModel columnModel) {
		if (containsKeyColumn(keyModel, columnModel)) {
			return;
		}
		ColumnRef ref = columnModel.getJiemamy().getFactory().newReference(columnModel);
		keyModel.getKeyColumns().add(ref);
	}
	
	/**
	 * キー制約の構成カラムに、指定したカラムが含まれるかどうか調べる。
	 * 
	 * @param keyModel 対象キー制約
	 * @param columnModel カラム
	 * @return 含まれる場合は{@code true}、そうでない場合は{@code false}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static boolean containsKeyColumn(KeyConstraint keyModel, ColumnModel columnModel) {
		Validate.notNull(keyModel);
		Validate.notNull(columnModel);
		
		List<ColumnRef> keyColumns = keyModel.getKeyColumns();
		for (ColumnRef columnRef : keyColumns) {
			if (columnModel.getId() == columnRef.getReferenceId()) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * キー制約の構成カラムから、指定したカラムを削除する。
	 * 
	 * @param keyModel 対象キー制約
	 * @param columnModel カラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static void removeKeyColumn(KeyConstraint keyModel, ColumnModel columnModel) {
		Validate.notNull(keyModel);
		Validate.notNull(columnModel);
		
		List<ColumnRef> keyColumns = keyModel.getKeyColumns();
		Iterator<ColumnRef> iterator = keyColumns.iterator();
		while (iterator.hasNext()) {
			ColumnRef columnRef = iterator.next();
			if (columnModel.getId() == columnRef.getReferenceId()) {
				iterator.remove();
			}
		}
	}
	
	/**
	 * キーカラム名のカンマ区切りの文字列に変換する。
	 * 
	 * @param keyModel 変換対象キー
	 * @return キーカラム名のカンマ区切りの文字列
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static String toStringKeyColumns(KeyConstraint keyModel) {
		Validate.notNull(keyModel);
		
		ReferenceResolver resolver = keyModel.getJiemamy().getReferenceResolver();
		List<ColumnRef> keyColumns = keyModel.getKeyColumns();
		List<String> columnNames = CollectionsUtil.newArrayList(keyColumns.size());
		for (ColumnRef columnRef : keyColumns) {
			columnNames.add(resolver.resolve(columnRef).getName());
		}
		return StringUtils.join(columnNames, ", ");
	}
	
	private KeyConstraintUtil() {
		throw new UnexpectedConditionError("不到達ポイント");
	}
}
