/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/21
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;
import java.util.Map;

import org.jiemamy.model.RootModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * エンティティ名が衝突しているかどうかを調べるバリデータ。
 * 
 * <ul>
 *   <li>{@link RootModel} が保持する全てのエンティティ間で、名前の重複があってはならない。</li>
 * </ul>
 * 
 * @author daisuke
 */
public class EntityNameCollisionValidator extends AbstractValidator {
	
	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		Collection<Problem> result = super.validate(rootModel);
		Map<String, Collection<EntityModel>> map = CollectionsUtil.newHashMap();
		
		for (EntityModel entityModel : rootModel.getEntities()) {
			String name = entityModel.getName();
			if (map.containsKey(name) == false) {
				map.put(name, new ArrayList<EntityModel>());
			}
			map.get(name).add(entityModel);
		}
		
		for (Map.Entry<String, Collection<EntityModel>> e : map.entrySet()) {
			if (e.getValue().size() != 1) {
				result.add(new EntityNameCollisionProblem(e.getKey(), e.getValue()));
			}
		}
		
		return result;
	}
	

	/**
	 * エンティティ名が衝突していることを示す{@link Problem}オブジェクト。
	 * 
	 * @author daisuke
	 */
	static class EntityNameCollisionProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param entityName 衝突したエンティティ名
		 * @param entities 衝突したエンティティの集合
		 */
		public EntityNameCollisionProblem(String entityName, Collection<EntityModel> entities) {
			super("E0070");
			registerMessage(Locale.JAPAN, "エンティティ名 \"{0}\" が重複しています");
			registerMessage(Locale.US, "Duplicate entity name {0}");
			setArguments(new Object[] {
				entityName
			});
			// TODO とりあえず実装。リソースバンドル使えた方がええよなぁ。
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
	}
}
